// ************* Version 02 ***************
// Node-Red
// funzioni https://www.packt.com/functions-arduino/
// https://github.com/doleron/esp8266-1-channel-relay-board-with-mqtt/blob/master/simple_mqtt/stc_15f104W.ino
// Dashboards https://stevesnoderedguide.com/node-red-dashboard
// Indicazioni di base per node-red https://github.com/pmmistry/Node-RED-Series/blob/master/Labs/lab_2.md
// antenna e misura segnale https://www.instructables.com/External-Antenna-for-ESP8266/
// Led di connessione wi-fi https://myesp8266.blogspot.com/2015/05/wifi-status-led.html
// voltmetro https://www.instructables.com/Voltmeter-Using-NodeMCU/

#include <PubSubClient.h>
#include <ESP8266WiFi.h>

//install status wifi LED on GPIO0
#define WIFI_LED_IO_MUX     PERIPHS_IO_MUX_GPIO0_U
#define WIFI_LED_IO_NUM     0
#define WIFI_LED_IO_FUNC    FUNC_GPIO0
// -------------------------------------------
// Wi-Fi access setup
#define WIFI_SSID "Uplink-Apple"    // Set SSID 
#define WIFI_PASSWORD "trillina"    // Set Wi-Fi Password
String newHostname = "PP-Distr-01"; // Set PowerPole Distributor hostname
// -------------------------------------------

// Variables declaration
const int Relay_1 = 4;   // D2 -> GPIO4
const int Relay_2 = 14;  // D5 -> GPIO14
const int Relay_3 = 12;  // D6 -> GPIO12
const int Relay_4 = 13;  // D7 -> GPIO13

unsigned long lastOperation;
const long MIN_OPERATION_INTERVAL = 1000L; // Define time between relay commands
int status = WL_IDLE_STATUS;
signed long now; // ***
String message; // ***

long lastMsg = 0; // Controls info publishing

int analogInput = A0; // Voltage measurement
float Vout = 0.00;
float Vin = 0.00;
float R1 = 46139.00;  // Write here real measured resistance R9 value
float R2 = 9957.00;   // Write here real measured resistance R10 value
int val = 0;

int IsBoot = 1; // Checks if it has been a BOOT
int HasBeenBoot = 1;
int IsConnected = 0; // handle Blue Led in the dashboard 

// -------------------------
float curr_SS;    // Current Wi-Fi signal strenght
String curr_IP;   // Current local IP
String curr_Host; // Current Hostname
String curr_fw_ver = "2_0"; // Current FW version
// -------------------------
void check_voltage(); // Voltage measurement function
void connect();       // Wi-Fi connection function
// -------------------------
// Set MQTT server configuration
char mqttServer[]  = "minilinux"; //change it for your MQTT server IP or network name // <--- ****************
int mqttPort  = 1883; //1883 is the default port for MQTT. Change if necessary
char deviceId[]  = "PP_Distr"; //every device should have a different name
char topic[]  = "/PP_Distr/command"; //the topic should be different for each device as well
char mqttUser[]  = "yourMQTTuser";
char mqttPassword[]  = "yourMQTTpassword";
// -------------------------
WiFiClient wifiClient;
PubSubClient client(wifiClient);

// ---------------------------------------
// -------------- SETUP ------------------
void setup() {

// delay(200); // Wait for 8 Vdc ready USELESS
            
// Set pin for Relays as outputs:
  pinMode(Relay_1, OUTPUT);
  pinMode(Relay_2, OUTPUT);
  pinMode(Relay_3, OUTPUT);
  pinMode(Relay_4, OUTPUT);

// Serial terminal activation
  Serial.begin(115200);
  delay(10);
  Serial.println("Let' start now");

  WiFi.mode(WIFI_STA);

}
// ---------------------------------------
// --------------- LOOP ------------------
void loop() {

// Info published only each 0,5 seconds
long sendMsg = millis(); 
if (sendMsg - lastMsg > 500) {
    lastMsg = sendMsg;    
// ----------------------------------------------------------------------
      // If BOOT dashboard buttons status reset
      if (IsBoot==1) {

          Serial.println("BOOT");
        
          // Force all relays Off at boot - Redudant but just to be sure that all relays are Off
          digitalWrite(Relay_1, LOW); 
          digitalWrite(Relay_2, LOW); 
          digitalWrite(Relay_3, LOW); 
          digitalWrite(Relay_4, LOW); 

        // Publish not dynamic info in topic: info/.........
        // Publish local IP
          curr_IP = WiFi.localIP().toString().c_str();  
          // curr_IP = "192.168.1.44"; // <--- **************** fixed for video making

          client.publish("info/wifi_IP", String(curr_IP).c_str(), true);
          // Publish Hostname
          curr_Host = WiFi.hostname();     
          client.publish("info/wifi_Host", String(curr_Host).c_str(), true); 
          // Publish current FW version
          client.publish("info/FW_ver", String(curr_fw_ver).c_str(), true); 

          IsBoot=0;
          HasBeenBoot=1;
      } else {
          // Serial.println("no BOOT");
      } 

    // Publishing dynamic info in topic: info/.........
        // Publish line voltage
        check_voltage(); 
        // client.publish("info/PS_Voltage", String(Vin,1).c_str(), true);
        client.publish("info/PS_Voltage", String(Vin).c_str(), true);
        Vin=0; // clear variable Vin (it will be used by dashboard to understand if distributor is fed or not)
        // Publish Wi-Fi signal strength
        curr_SS = WiFi.RSSI();
        client.publish("info/wifi_signal", String(curr_SS).c_str(), true);

        // Publish connection status 
        if (IsConnected == 1) {
            client.publish("info/wifi_connected", String(1).c_str(), true); // Connected
        } else {
            client.publish("info/wifi_connected", String(0).c_str(), true); // Disonnected 
        }

    } 
// --------------------------------------------------------
  if ( !client.connected() ) {
    connect();
  }
  client.loop();
}
// ---------------------------------------
// MQTT commands management (the nodeMCU stays listening on topic)
void callback(char* topic, byte* payload, unsigned int length) {

    // Inhibits commands if time between commands is < MIN_OPERATION_INTERVAL
    signed long now = millis();
    long deltaTime = now - lastOperation;
    if (deltaTime > MIN_OPERATION_INTERVAL) {
      String message = ""; // *** 
      for (int i = 0; i < length; i++) {
        message = message + (char)payload[i];
      }

  // Managing incoming commands (topic: relay_command/.........) and status LED (topic: relay_status/.........)
      // --------- Relay 1 ---------------
      if(message == "OPEN_1") { 
        Serial.println("Performing open command - "+message);
        digitalWrite(Relay_1, LOW);
        client.publish("relay_status/RELAY_State_1", String(0).c_str());
        lastOperation = now;
      } else if(message == "CLOSE_1") {
         Serial.println("Performing close command - "+message);
         digitalWrite(Relay_1, HIGH);
         client.publish("relay_status/RELAY_State_1", String(1).c_str());
        lastOperation = now;
      }
      // --------- Relay 2 ---------------
      if(message == "OPEN_2") { 
        Serial.println("Performing open command - "+message);
        digitalWrite(Relay_2, LOW);
        client.publish("relay_status/RELAY_State_2", String(0).c_str());
        lastOperation = now;
      } else if(message == "CLOSE_2") {
        Serial.println("Performing close command - "+message);
        digitalWrite(Relay_2, HIGH);
        client.publish("relay_status/RELAY_State_2", String(1).c_str());
        lastOperation = now;
      }

      // --------- Relay 3 ---------------
      if(message == "OPEN_3") { 
        Serial.println("Performing open command - "+message);
        digitalWrite(Relay_3, LOW);
        client.publish("relay_status/RELAY_State_3", String(0).c_str());
        lastOperation = now;
      } else if(message == "CLOSE_3") {
        Serial.println("Performing close command - "+message);
        digitalWrite(Relay_3, HIGH);
        client.publish("relay_status/RELAY_State_3", String(1).c_str());
        lastOperation = now;
      }

      // --------- Relay 4 ---------------
      if(message == "OPEN_4") { 
        Serial.println("Performing open command - "+message);
        digitalWrite(Relay_4, LOW);
        client.publish("relay_status/RELAY_State_4", String(0).c_str());
        lastOperation = now;
      } else if(message == "CLOSE_4") {
        Serial.println("Performing close command - "+message);
        digitalWrite(Relay_4, HIGH);
        client.publish("relay_status/RELAY_State_4", String(1).c_str());
        lastOperation = now;
      }
      // ------------------------
  } else {
        // Sends WAIT message to dashboard if time between commands is < MIN_OPERATION_INTERVAL
        if (HasBeenBoot==0) {
            Serial.println("Operation denied right now. Try again later.");
            // Print message WAIT on dashboard
            client.publish("info/pse_wait", String(0).c_str(), true); 
            delay(MIN_OPERATION_INTERVAL+500);
            client.publish("info/pse_wait", String(1).c_str(), true); 
        } else {
            client.publish("info/pse_wait", String(1).c_str(), true); 
            HasBeenBoot=0;
        }
  }
}
// ---------------------------------------
void connect() {
  
  while (!client.connected()) {
    
    status = WiFi.status();
    // Connection status blue LED activation
    wifi_status_led_install(WIFI_LED_IO_NUM, WIFI_LED_IO_MUX, FUNC_GPIO0); 

    if ( status != WL_CONNECTED) {
      WiFi.hostname(newHostname.c_str()); // set hostname
      WiFi.begin(WIFI_SSID, WIFI_PASSWORD);
      while (WiFi.status() != WL_CONNECTED) {
        delay(500);
        Serial.print(".");
      }
      Serial.println();
      Serial.print("Connected to ");
      Serial.println(WIFI_SSID);
      Serial.print("IP address: "); 
      Serial.println(WiFi.localIP());
      Serial.print("Hostname: "); 
      Serial.println(WiFi.hostname());

      // Keep blue led active until connected
      digitalWrite(WIFI_LED_IO_NUM,HIGH); // physical Blue LED on the unit
      IsConnected = 1;
      // client.publish("info/wifi_connected", String(1).c_str(), true); // Connected LED on the dashboard

      // print signal strenght
      Serial.println("------------");
      Serial.println("Signal strenght [dBm]");
      Serial.println(WiFi.RSSI());
      Serial.println("------------");

    }
    // Connection to MQTT server
    client.setServer(mqttServer, mqttPort);
    client.setCallback(callback);
    if (client.connect(deviceId, mqttUser, mqttPassword)) {
      client.subscribe(topic);
      Serial.println("Connected to MQTT Server");
    } else {
      Serial.print("[FAILED] [ rc = ");
      Serial.print(client.state() );
      Serial.println(" : retrying in 5 seconds]");
      delay(5000);
    }
  }
}
// ---------------------------------------
void check_voltage() {
 val = analogRead(analogInput);//reads the analog input
    Vout = (val * 3.3) / 1024.00; // formula for calculating voltage read by A0 pin nodeMCU 3.3
    Vout = Vout * 0.9745; // 0.9745 EMPIRICAL corrective coefficient to match real to shown voltage value
    Vin = Vout / (R2/(R1+R2)); // formula for calculating voltage in i.e. GND   
    // If no incoming line active voltage measurement forced to 0
    
    if (Vin<10){
      Vin=0;
    }   
    // Serial.print(Vin);
}
